"""
SyNumSeS
========

SyNumSeS is a Python package for simulating various semiconductor devices
in one dimension.
It solves the the *Van Roosbroeck equations* using the *Scharfetter Gummel Discretization*.

"""


#import synumses.one_dimension
from .one_dimension import *
